"use strict";
/*!
 * Contentstack DataSync Manager
 * Copyright (c) 2019 Contentstack LLC
 * MIT Licensed
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.unlock = exports.lock = exports.poke = exports.pop = exports.unshift = exports.push = exports.init = void 0;
const debug_1 = __importDefault(require("debug"));
const events_1 = require("events");
const lodash_1 = require("lodash");
const __1 = require("../");
const api_1 = require("../api");
const fs_1 = require("../util/fs");
const index_1 = require("../util/index");
const logger_1 = require("../util/logger");
const promise_map_1 = require("../util/promise.map");
const inet_1 = require("./inet");
const q_1 = require("./q");
const token_management_1 = require("./token-management");
const debug = (0, debug_1.default)('sync-core');
const emitter = new events_1.EventEmitter();
const formattedAssetType = '_assets';
const formattedContentType = '_content_types';
const flag = {
    SQ: false,
    WQ: false,
    lockdown: false,
};
let config;
let Contentstack;
let Q;
/**
 * @description Core's primary. This is where it starts..
 * @param {Object} connector - Content connector instance
 * @param {Object} config - Application config
 */
const init = (contentStore, assetStore) => {
    config = (0, __1.getConfig)();
    Q = new q_1.Q(contentStore, assetStore, config);
    (0, api_1.init)(config.contentstack);
    debug('Sync core:start invoked');
    return new Promise((resolve, reject) => {
        try {
            Contentstack = config.contentstack;
            const paths = config.paths;
            const environment = Contentstack.environment || process.env.NODE_ENV || 'development';
            debug(`Environment: ${environment}`);
            const request = {
                qs: {
                    environment,
                    limit: config.syncManager.limit,
                },
            };
            if (typeof Contentstack.sync_token === 'string' && Contentstack.sync_token.length !== 0) {
                request.qs.sync_token = Contentstack.sync_token;
            }
            else if (typeof Contentstack.pagination_token === 'string' && Contentstack.pagination_token.length !== 0) {
                request.qs.pagination_token = Contentstack.pagination_token;
            }
            else if ((0, fs_1.existsSync)(paths.token)) {
                const token = JSON.parse((0, fs_1.readFileSync)(paths.token));
                request.qs[token.name] = token.token;
            }
            else {
                request.qs.init = true;
                if (config.syncManager.filters && typeof config.syncManager.filters === 'object') {
                    const filters = config.syncManager.filters;
                    // tslint:disable-next-line: forin
                    for (const filter in filters) {
                        request.qs[filter] = filters[filter].join(',');
                    }
                }
            }
            return fire(request)
                .then(resolve)
                .catch(reject);
        }
        catch (error) {
            return reject(error);
        }
    });
};
exports.init = init;
const push = (data) => {
    Q.emit('push', data);
};
exports.push = push;
const unshift = (data) => {
    Q.emit('push', data);
};
exports.unshift = unshift;
const pop = () => {
    Q.emit('pop');
};
exports.pop = pop;
/**
 * @description Notifies the sync manager utility to wake up and start syncing..
 */
const poke = () => __awaiter(void 0, void 0, void 0, function* () {
    try {
        debug('Invoked poke');
        logger_1.logger.info('Received \'contentstack sync\' notification');
        if (!flag.lockdown) {
            flag.WQ = true;
            return yield check();
        }
        return null;
    }
    catch (error) {
        debug('Error [poke]', error);
        throw error;
    }
});
exports.poke = poke;
/**
 * @description Check's if the status of the app when a new incoming notification is fired
 * @description Starts processing if the 'SQ: false'
 */
const check = () => __awaiter(void 0, void 0, void 0, function* () {
    try {
        debug(`Check called. SQ status is ${flag.SQ} and WQ status is ${flag.WQ}`);
        if (!flag.SQ && flag.WQ) {
            flag.WQ = false;
            flag.SQ = true;
            yield sync();
            debug(`Sync completed and SQ flag updated. Cooloff duration is ${config.syncManager.cooloff}`);
            setTimeout(() => {
                flag.SQ = false;
                emitter.emit('check');
            }, config.syncManager.cooloff);
        }
    }
    catch (error) {
        logger_1.logger.error(error);
        debug('Error [check]', error);
        check().then(() => {
            debug('passed [check] error');
        }).catch((error) => {
            debug('failed [check] error', error);
        });
        throw error;
    }
});
/**
 * @description Gets saved token, builds request object and fires the sync process
 */
const sync = () => __awaiter(void 0, void 0, void 0, function* () {
    try {
        debug('started [sync]');
        const tokenObject = yield (0, token_management_1.getToken)();
        debug('tokenObject [sync]', tokenObject);
        const token = tokenObject;
        const request = {
            qs: {
                environment: process.env.SYNC_ENV || Contentstack.environment || 'development',
                limit: config.syncManager.limit,
                [token.name]: token.token,
            },
        };
        return yield fire(request);
    }
    catch (error) {
        debug('Error [sync]', error);
        throw error;
    }
});
/**
 * @description Used to lockdown the 'sync' process in case of exceptions
 */
const lock = () => {
    debug('Contentstack sync locked..');
    flag.lockdown = true;
};
exports.lock = lock;
/**
 * @description Used to unlock the 'sync' process in case of errors/exceptions
 */
const unlock = (refire) => {
    debug('Contentstack sync unlocked..', refire);
    flag.lockdown = false;
    if (typeof refire === 'boolean' && refire) {
        flag.WQ = true;
        if (flag.requestCache && Object.keys(flag.requestCache)) {
            return fire(flag.requestCache.params)
                .then(flag.requestCache.resolve)
                .catch(flag.requestCache.reject);
        }
    }
    check();
};
exports.unlock = unlock;
/**
 * @description Description required
 * @param {Object} req - Contentstack sync API request object
 */
const fire = (req) => {
    debug(`Fire called with: ${JSON.stringify(req)}`);
    flag.SQ = true;
    return new Promise((resolve, reject) => {
        return (0, api_1.get)(req).then((response) => {
            delete req.qs.init;
            delete req.qs.pagination_token;
            delete req.qs.sync_token;
            delete req.path;
            const syncResponse = response;
            debug('Response [fire]', syncResponse.items.length);
            if (syncResponse.items.length) {
                return (0, index_1.filterItems)(syncResponse, config).then(() => {
                    if (syncResponse.items.length === 0) {
                        return postProcess(req, syncResponse)
                            .then(resolve)
                            .catch(reject);
                    }
                    syncResponse.items = (0, index_1.formatItems)(syncResponse.items, config);
                    let groupedItems = (0, index_1.groupItems)(syncResponse.items);
                    groupedItems = (0, index_1.markCheckpoint)(groupedItems, syncResponse);
                    // send assets data for processing
                    if (groupedItems[formattedAssetType]) {
                        groupedItems[formattedAssetType].forEach((asset) => {
                            Q.push(asset);
                        });
                        delete groupedItems[formattedAssetType];
                    }
                    if (groupedItems[formattedContentType]) {
                        groupedItems[formattedContentType].forEach((contentType) => {
                            Q.push(contentType);
                        });
                        delete groupedItems[formattedContentType];
                    }
                    const contentTypeUids = Object.keys(groupedItems);
                    (0, lodash_1.remove)(contentTypeUids, (contentTypeUid) => {
                        const contentType = groupedItems[contentTypeUid];
                        if (contentType.length === 1 && !contentType[0].publish_details) {
                            Q.push(contentType[0]);
                            return true;
                        }
                        return false;
                    });
                    return (0, promise_map_1.map)(contentTypeUids, (uid) => {
                        return new Promise((mapResolve, mapReject) => {
                            return (0, api_1.get)({
                                path: `${Contentstack.apis.content_types}${uid}`,
                                qs: {
                                    include_global_field_schema: config.contentstack.query.include_global_field_schema,
                                },
                            }).then((contentTypeSchemaResponse) => {
                                const schemaResponse = contentTypeSchemaResponse;
                                if (schemaResponse.content_type) {
                                    const items = groupedItems[uid];
                                    items.forEach((entry) => {
                                        entry._content_type = (0, lodash_1.cloneDeep)(schemaResponse.content_type);
                                        Q.push(entry);
                                    });
                                    return mapResolve('');
                                }
                                const err = new Error('Content type ${uid} schema not found!');
                                // Illegal content type call
                                err.code = 'ICTC';
                                return mapReject(err);
                            }).catch((error) => {
                                if ((0, inet_1.netConnectivityIssues)(error)) {
                                    flag.SQ = false;
                                }
                                return mapReject(error);
                            });
                        });
                    }, 2).then(() => {
                        return postProcess(req, syncResponse)
                            .then(resolve)
                            .catch(reject);
                    }).catch((error) => {
                        if ((0, inet_1.netConnectivityIssues)(error)) {
                            flag.SQ = false;
                        }
                        // Errorred while fetching content type schema
                        return reject(error);
                    });
                }).catch((processError) => {
                    return reject(processError);
                });
            }
            return postProcess(req, syncResponse)
                .then(resolve);
        }).catch((error) => {
            debug('Error [fire]', error);
            if ((0, inet_1.netConnectivityIssues)(error)) {
                flag.SQ = false;
            }
            // do something
            return reject(error);
        });
    });
};
/**
 * @description Process data once 'sync' data has been fetched
 * @param {Object} req - Sync API request object
 * @param {Object} resp - Sync API response object
 */
const postProcess = (req, resp) => {
    return new Promise((resolve, reject) => {
        let name;
        if (resp.pagination_token) {
            name = 'pagination_token';
        }
        else {
            name = 'sync_token';
        }
        return (0, token_management_1.saveCheckpoint)(name, resp[name])
            .then(() => {
            // re-fire!
            req.qs[name] = resp[name];
            if (flag.lockdown) {
                logger_1.logger.log('Checkpoint: lockdown has been invoked');
                flag.requestCache = {
                    params: req,
                    reject,
                    resolve,
                };
            }
            else {
                if (name === 'sync_token') {
                    flag.SQ = false;
                    return resolve('');
                }
                return fire(req)
                    .then(resolve)
                    .catch(reject);
            }
        })
            .catch(reject);
    });
};
emitter.on('check', check);
